#include "fastSimplex.hh"

#include <iostream>
#include <sstream>
#include <string>

#include <stdlib.h>

#define RUN_TRACKER

#ifdef CHECK_WITH_PPL
#include <ppl.hh>

namespace PPL=Parma_Polyhedra_Library;
#endif

struct Constraint {
  LinearForm<coefficient> form;
  bool isStrict;
  coefficient constant;

  Constraint(const std::vector<std::string>& variableNames, bool isStrict0, const coefficient& constant0) :
    form(variableNames), isStrict(isStrict0), constant(constant0) {
  }
};

std::ostream& operator<<(std::ostream& out, const Constraint& constraint) {
  return (out << "(" << (constraint.isStrict ? "< " : "<= ") << constraint.form << constraint.constant << ")");
}

struct TestCase {
  std::vector<std::string> variableNames;
  std::vector<Constraint> constraints;
  std::vector<unsigned> blockSizes;
  unsigned numberOfTrueVariables;

public:
  TestCase(unsigned numberOfVariables, unsigned numberOfConstraints, unsigned numberOfConstraintBlocks, unsigned coefficientMagnitude, double nonZeroProbability);

  Constraint randomConstraint();
  void test();
};

static inline std::string trueVariableName(unsigned i) {
  static const char alphabet[] = "abcdefghijklmnopqrstuvwxyz";
  const unsigned nLetters = (sizeof(alphabet)/sizeof(alphabet[0]))-1;

  std::ostringstream s;
  s << (alphabet[i % nLetters]);
  if (i >= nLetters) {
    s << (i / nLetters);
  }
  return s.str();
}

static inline std::string deltaVariableName(unsigned i) {
  std::ostringstream s;
  s << "delta" << (i+1);
  return s.str();
}

static inline unsigned getRandom(unsigned magnitude) {
  return ((unsigned) rand()) % magnitude;
}

static inline coefficient randomCoefficient(unsigned magnitude) {
  unsigned r = getRandom(magnitude << 1);
  if ((r & 1) != 0) {
    return -coefficient(int (r>>1));
  } else {
    return coefficient(int (r>>1));
  }
}

TestCase::TestCase(unsigned numberOfVariables, unsigned numberOfConstraints, unsigned numberOfConstraintBlocks, unsigned coefficientMagnitude, double nonZeroProbability) :
  numberOfTrueVariables(numberOfVariables)
{
  {
    assert(numberOfConstraintBlocks > 0);
    unsigned remaining = numberOfConstraints;

    for(unsigned i=0; i<numberOfConstraintBlocks-1; i++) {
      unsigned blockSize =
	std::min(getRandom(1+remaining-(numberOfConstraintBlocks-i)), (remaining-1));

      blockSizes.push_back(blockSize);
      remaining -= blockSize;
    }
    blockSizes.push_back(remaining);
  }

  for(unsigned i=0; i<numberOfVariables; i++) {
    variableNames.push_back(trueVariableName(i));
  }
  for(unsigned i=0; i<numberOfConstraints; i++) {
    variableNames.push_back(deltaVariableName(i));
  }
  for(unsigned i=0; i<numberOfConstraints; i++) {
    Constraint constraint(variableNames, (rand() & 1)!=0, randomCoefficient(coefficientMagnitude));
    bool hasNonZero = false;
    for(unsigned j=0; j<numberOfVariables; j++) {
      if (nonZeroProbability > 0 && (nonZeroProbability>=1 || rand() < RAND_MAX*nonZeroProbability)) {
	coefficient coeff=randomCoefficient(coefficientMagnitude);
	if (coeff != 0) {
	  constraint.form.addMonomial(j, coeff);
	  hasNonZero = true;
	}
      }
    }
    if (!hasNonZero) {
      constraint.form.addMonomial(rand() % numberOfVariables, 42);
    }
    constraints.push_back(constraint);
  }
}

std::ostream& operator<<(std::ostream& out, const TestCase& testCase) {
  unsigned i=0;
  out << testCase.blockSizes.size() << " constraint blocks" << std::endl;
  for(unsigned block=0; block<testCase.blockSizes.size(); block++) {
    assert(i < testCase.constraints.size());
    out << "{ ";
    for(unsigned j=0; j<testCase.blockSizes[block]; j++) {
      if (j > 0) out << "," << std::endl;
      out << "  " << testCase.constraints[i];
      i++;
    }
    out << " }" << std::endl;
  }
  assert(i == testCase.constraints.size());
  return out;
}

#ifdef CHECK_WITH_PPL
static PPL::Constraint toPPLConstraint(const Constraint& constraint) {
  PPL::Linear_Expression expr;
  for(unsigned i=0; i<constraint.form.size(); i++) {
    expr += mpq_class(constraint.form[i]) * PPL::Variable(i);
  }
  if (constraint.isStrict) {
    return (expr < mpq_class(constraint.constant));
  } else {
    return (expr <= mpq_class(constraint.constant));
  }
}
#endif

void TestCase::test() {
#ifdef CHECK_WITH_PPL
  PPL::NNC_Polyhedron poly(variableNames.size());
#endif
#ifdef RUN_TRACKER
  MY_SIMPLEX simplex(variableNames, numberOfTrueVariables, true);
#else
  MY_SIMPLEX simplex(variableNames, numberOfTrueVariables, false);
#endif

  for(unsigned i=0; i<constraints.size(); i++) {
    simplex.assignBasic(i+numberOfTrueVariables, constraints[i].form);
  }

  unsigned constraint=0;
  for(unsigned block=0; block<blockSizes.size(); block++) {
#ifdef CHECK_WITH_PPL
    PPL::Constraint_System cs;
#endif
    for(unsigned i=0; i<blockSizes[block]; i++) {
      const Constraint& ct = constraints[constraint];
#ifdef CHECK_WITH_PPL
      cs.insert(toPPLConstraint(ct));
#endif
      if (ct.isStrict) {
	simplex.assertUpperBound(constraint+numberOfTrueVariables, coordinate(ct.constant, -1));
      } else {
	simplex.assertUpperBound(constraint+numberOfTrueVariables, coordinate(ct.constant));
      }
      constraint++;
    }
#ifdef CHECK_WITH_PPL
    poly.add_constraints(cs);
#endif

    bool answerSimplex=simplex.check();
#ifdef CHECK_WITH_PPL
    bool answerPoly= !poly.is_empty();
    if (answerPoly != answerSimplex) {
      std::cout << "ERROR! PPL: " << std::boolalpha << answerPoly << " SIMPLEX: " << answerSimplex << std::endl;
    } else {
      std::cout << std::boolalpha << answerSimplex << " " << answerPoly << std::endl;
    }
#else
    std::cout << std::boolalpha << answerSimplex << std::endl;
#endif
  }
  simplex.printStatistics(std::cerr);
}

int main(int argc, char **argv) {
  if (argc != 6 && argc != 7) {
    exit(2);
  }

  unsigned numberOfCases = atoi(argv[1]), 
    numberOfVariables = atoi(argv[2]),
    numberOfConstraints = atoi(argv[3]),
    numberOfBlocks = atoi(argv[4]),
    coefficientSize = atoi(argv[5]);
  double nonZeroProbability = (argc == 7) ? atof(argv[6]) : 1.0;

  for(unsigned i=0; i<numberOfCases; i++) {
    //TestCase testCase(8, 30, 30, 100);
    TestCase testCase(numberOfVariables, numberOfConstraints, numberOfBlocks, coefficientSize, nonZeroProbability);
    //std::cout << testCase << std::endl;
    testCase.test();
  }
}
